@@ -46,15 +46,19 @@ typedef struct _OutputQueue
   GQueue        *queue;
 } OutputQueue;
 
+typedef void (*PushedCb) (OutputQueue *q, gpointer user_data, GError *error);
+
 typedef struct _OutputQueueElem
 {
   OutputQueue  *queue;
   const guint8 *buf;
   gsize         size;
-  GFunc         cb;
+  PushedCb      cb;
   gpointer      user_data;
 } OutputQueueElem;
 
+static GCancellable *cancel;
+
 static OutputQueue*
 output_queue_new (GOutputStream *output)
 {
@@ -90,6 +94,8 @@ output_queue_ref (OutputQueue *q)
 static void
 output_queue_unref (OutputQueue *q)
 {
+  g_return_if_fail (q != NULL);
+
   q->refs--;
   if (q->refs == 0)
     output_queue_free (q);
@@ -143,26 +149,21 @@ output_queue_idle (gpointer user_data)
     }
 
   g_debug ("flushing %" G_GSIZE_FORMAT, e->size);
-  g_output_stream_write_all (q->output, e->buf, e->size, NULL, NULL, &error);
-  if (error)
-    goto end;
-
+  g_output_stream_write_all (q->output, e->buf, e->size, NULL, cancel, &error);
   if (e->cb)
-    e->cb (q, e->user_data);
+    e->cb (q, e->user_data, error);
+
+  if (error)
+      goto end;
 
   q->flushing = TRUE;
-  g_output_stream_flush_async (q->output, G_PRIORITY_DEFAULT, NULL, output_queue_flush_cb, e);
+  g_output_stream_flush_async (q->output, G_PRIORITY_DEFAULT, cancel, output_queue_flush_cb, e);
 
   q->idle_id = 0;
   return FALSE;
 
 end:
-  if (error)
-    {
-      g_warning ("error: %s", error->message);
-      g_clear_error (&error);
-    }
-
+  g_clear_error (&error);
   q->idle_id = 0;
   g_free (e);
   output_queue_unref (q);
@@ -172,10 +173,13 @@ end:
 
 static void
 output_queue_push (OutputQueue *q, const guint8 *buf, gsize size,
-                   GFunc pushed_cb, gpointer user_data)
+                   PushedCb pushed_cb, gpointer user_data)
 {
-  OutputQueueElem *e = g_new (OutputQueueElem, 1);
+  OutputQueueElem *e;
 
+  g_return_if_fail (q != NULL);
+
+  e = g_new (OutputQueueElem, 1);
   e->buf = buf;
   e->size = size;
   e->cb = pushed_cb;
@@ -204,11 +208,19 @@ typedef struct _Client
   OutputQueue       *queue;
 } Client;
 
+static gboolean quit_service;
 static GMainLoop *loop;
 static GInputStream *mux_istream;
 static GOutputStream *mux_ostream;
 static OutputQueue *mux_queue;
 static GHashTable *clients;
+static GSocketService *socket_service;
+#ifdef G_OS_UNIX
+static gint port_fd;
+#elif defined(G_OS_WIN32)
+static HANDLE port_handle;
+#endif
+
 static void start_mux_read (GInputStream *istream);
 
 static void
@@ -216,6 +228,9 @@ quit (int sig)
 {
   g_debug ("quit %d", sig);
 
+  if (sig == SIGINT || sig == SIGTERM)
+      quit_service = TRUE;
+
   g_main_loop_quit (loop);
 }
 
@@ -246,6 +261,9 @@ add_client (GSocketConnection *client_connection)
 static void
 client_free (Client *c)
 {
+  g_debug ("Free client %p", c);
+
+  g_io_stream_close (G_IO_STREAM (c->client_connection), NULL, NULL);
   g_object_unref (c->client_connection);
   output_queue_unref (c->queue);
   g_free (c);
@@ -256,7 +274,7 @@ remove_client (Client *client)
 {
   g_debug ("remove client %p", client);
 
-  g_hash_table_remove (clients, client->client_connection);
+  g_hash_table_remove (clients, &client->id);
 }
 
 typedef struct ReadData
@@ -283,9 +301,9 @@ read_thread (GSimpleAsyncResult *simple,
                            data->buffer, data->count, &bread,
                            cancellable, &error);
   g_debug ("my read result %" G_GSIZE_FORMAT, bread);
-  if (bread != data->count)
+  if (bread != data->count) {
     data->size = -1;
-  else
+  } else
     data->size = bread;
 
   if (error)
@@ -343,8 +361,26 @@ my_input_stream_read_finish (GInputStream *stream,
 }
 
 static void
-mux_pushed_client_cb (OutputQueue *q, gpointer user_data)
+handle_push_error (OutputQueue *q, gpointer user_data, GError *error)
 {
+  Client *client = user_data;
+
+  if (!error)
+    return;
+
+  g_warning ("push error: %s", error->message);
+  remove_client (client);
+  return;
+}
+
+static void
+mux_pushed_client_cb (OutputQueue *q, gpointer user_data, GError *error)
+{
+  if (error) {
+    handle_push_error (q, user_data, error);
+    return;
+  }
+
   start_mux_read (mux_istream);
 }
 
@@ -362,17 +398,17 @@ mux_data_read_cb (GObject      *source_object,
     {
       g_warning ("error: %s", error->message);
       g_clear_error (&error);
-      quit (1);
+      quit (-1);
       return;
     }
 
-  g_debug ("looking up client %" G_GINT64_FORMAT, demux.client);
   Client *c = g_hash_table_lookup (clients, &demux.client);
+  g_debug ("looked up client: %p", c);
   g_warn_if_fail(c != NULL);
 
   if (c)
     output_queue_push (c->queue, (guint8 *) demux.buf, demux.size,
-                       (GFunc) mux_pushed_client_cb, c);
+                       mux_pushed_client_cb, c);
   else
     start_mux_read (mux_istream);
 }
@@ -392,7 +428,7 @@ mux_size_read_cb (GObject      *source_object,
 
   my_input_stream_read_async (istream,
                               &demux.buf, demux.size, G_PRIORITY_DEFAULT,
-                              NULL, mux_data_read_cb, NULL);
+                              cancel, mux_data_read_cb, NULL);
   return;
 
 end:
@@ -402,7 +438,7 @@ end:
       g_clear_error (&error);
     }
 
-  quit (2);
+  quit (-2);
 }
 
 static void
@@ -420,7 +456,7 @@ mux_client_read_cb (GObject      *source_object,
     goto end;
   my_input_stream_read_async (istream,
                               &demux.size, sizeof (guint16), G_PRIORITY_DEFAULT,
-                              NULL, mux_size_read_cb, NULL);
+                              cancel, mux_size_read_cb, NULL);
   return;
 
 end:
@@ -430,7 +466,7 @@ end:
       g_clear_error (&error);
     }
 
-  quit (3);
+  quit (-3);
 }
 
 static void
@@ -438,16 +474,21 @@ start_mux_read (GInputStream *istream)
 {
   my_input_stream_read_async (istream,
                               &demux.client, sizeof (gint64), G_PRIORITY_DEFAULT,
-                              NULL, mux_client_read_cb, NULL);
+                              cancel, mux_client_read_cb, NULL);
 }
 
 static void client_start_read (Client *client);
 
 static void
-mux_pushed_cb (OutputQueue *q, gpointer user_data)
+mux_pushed_cb (OutputQueue *q, gpointer user_data, GError *error)
 {
   Client *client = user_data;
 
+  if (error) {
+      handle_push_error (q, client, error);
+      return;
+  }
+
   if (client->size == 0)
     {
       remove_client (client);
@@ -467,7 +508,7 @@ client_read_cb (GObject      *source_object,
   gssize size;
 
   size = g_input_stream_read_finish (G_INPUT_STREAM (source_object), res, &error);
-  g_debug ("end read %" G_GSIZE_FORMAT, size);
+  g_debug ("end read %" G_GSSIZE_FORMAT, size);
   if (error)
     {
       g_warning ("error: %s", error->message);
@@ -480,9 +521,9 @@ client_read_cb (GObject      *source_object,
   g_return_if_fail (size >= 0);
   client->size = size;
 
-  output_queue_push (mux_queue, (guint8 *) &client->id, sizeof (gint64), NULL, NULL);
-  output_queue_push (mux_queue, (guint8 *) &client->size, sizeof (guint16), NULL, NULL);
-  output_queue_push (mux_queue, (guint8 *) client->buf, size, (GFunc) mux_pushed_cb, client);
+  output_queue_push (mux_queue, (guint8 *) &client->id, sizeof (gint64), handle_push_error, client);
+  output_queue_push (mux_queue, (guint8 *) &client->size, sizeof (guint16), handle_push_error, client);
+  output_queue_push (mux_queue, (guint8 *) client->buf, size, mux_pushed_cb, client);
 
   return;
 }
@@ -493,7 +534,7 @@ client_start_read (Client *client)
   GIOStream *iostream = G_IO_STREAM (client->client_connection);
   GInputStream *istream = g_io_stream_get_input_stream (iostream);
 
-  g_debug ("start read");
+  g_debug ("start read client %p", client);
   g_input_stream_read_async (istream,
                              client->buf, G_MAXUINT16, G_PRIORITY_DEFAULT,
                              NULL, client_read_cb, client);
@@ -643,32 +684,89 @@ open_mux_path (const char *path)
   g_return_if_fail (!mux_ostream);
   g_return_if_fail (!mux_queue);
 
+  g_debug ("Open %s", path);
 #ifdef G_OS_UNIX
-  gint fd = g_open (path, O_RDWR);
-  if (fd == -1)
+  port_fd = g_open (path, O_RDWR);
+  if (port_fd == -1)
       exit (1);
 
-  wait_for_virtio_host (fd);
+  wait_for_virtio_host (port_fd);
 
-  mux_ostream = g_unix_output_stream_new (fd, TRUE);
-  mux_istream = g_unix_input_stream_new (fd, FALSE);
+  mux_ostream = g_unix_output_stream_new (port_fd, TRUE);
+  mux_istream = g_unix_input_stream_new (port_fd, FALSE);
 #else
-  HANDLE h = CreateFile (path,
+  port_handle = CreateFile (path,
                          GENERIC_WRITE | GENERIC_READ,
                          0,
                          NULL,
                          OPEN_EXISTING,
                          FILE_FLAG_OVERLAPPED,
                          NULL);
-  g_assert (h != INVALID_HANDLE_VALUE);
+  g_assert (port_handle != INVALID_HANDLE_VALUE);
 
-  mux_ostream = G_OUTPUT_STREAM (g_win32_output_stream_new (h, TRUE));
-  mux_istream = G_INPUT_STREAM (g_win32_input_stream_new (h, TRUE));
+  mux_ostream = G_OUTPUT_STREAM (g_win32_output_stream_new (port_handle, TRUE));
+  mux_istream = G_INPUT_STREAM (g_win32_input_stream_new (port_handle, TRUE));
 #endif
 
   mux_queue = output_queue_new (G_OUTPUT_STREAM (mux_ostream));
+}
+
+static void
+run_service (void)
+{
+  g_debug ("Run service");
+
+  g_socket_service_start (socket_service);
+
+  cancel = g_cancellable_new ();
+  clients = g_hash_table_new_full (g_int64_hash, g_int64_equal,
+                                   NULL, (GDestroyNotify) client_free);
+
+  loop = g_main_loop_new (NULL, TRUE);
+#ifdef G_OS_UNIX
+  open_mux_path ("/dev/virtio-ports/org.spice-space.webdav.0");
+#else
+  open_mux_path ("\\\\.\\Global\\org.spice-space.webdav.0");
+#endif
+
+  /* listen on port for incoming clients, multiplex there input into
+     virtio path, demultiplex input from there to the respective
+     clients */
+
+#ifdef WITH_AVAHI
+  GError *error = NULL;
+
+  mdns_client = ga_client_new (GA_CLIENT_FLAG_NO_FLAGS);
+  g_signal_connect (mdns_client, "state-changed", G_CALLBACK (mdns_state_changed), NULL);
+  if (!ga_client_start (mdns_client, &error))
+    {
+      g_printerr ("%s\n", error->message);
+      exit (1);
+    }
+#endif
 
   start_mux_read (mux_istream);
+  g_main_loop_run (loop);
+  g_main_loop_unref (loop);
+
+  g_cancellable_cancel (cancel);
+
+  g_clear_object (&mux_istream);
+  g_clear_object (&mux_ostream);
+
+  output_queue_unref (mux_queue);
+  g_hash_table_unref (clients);
+
+  g_socket_service_stop (socket_service);
+
+  mux_queue = NULL;
+  g_clear_object (&cancel);
+
+#ifdef G_OS_WIN32
+  CloseHandle (port_handle);
+#else
+  close (port_fd);
+#endif
 }
 
 #ifdef G_OS_WIN32
@@ -718,7 +816,10 @@ service_main (DWORD argc, TCHAR *argv[])
   service_status.dwWaitHint = 0;
   SetServiceStatus (service_status_handle, &service_status);
 
-  g_main_loop_run (loop);
+  while (!quit_service) {
+      run_service ();
+      g_usleep (G_USEC_PER_SEC);
+  }
 
   service_status.dwCurrentState = SERVICE_STOPPED;
   SetServiceStatus (service_status_handle, &service_status);
@@ -757,13 +858,16 @@ main (int argc, char *argv[])
 
   signal (SIGINT, quit);
 
-
-  GSocketService *service = g_socket_service_new ();
+  /* run socket service once at beginning, there seems to be a bug on
+     windows, and it can't accept new connections if cleanup and
+     restart a new service */
+  socket_service = g_socket_service_new ();
   GInetAddress *iaddr = g_inet_address_new_any (G_SOCKET_FAMILY_IPV4);
   GSocketAddress *saddr = g_inet_socket_address_new (iaddr, port);
   g_object_unref (iaddr);
 
-  g_socket_listener_add_address (G_SOCKET_LISTENER (service), saddr,
+  g_socket_listener_add_address (G_SOCKET_LISTENER (socket_service),
+                                 saddr,
                                  G_SOCKET_TYPE_STREAM,
                                  G_SOCKET_PROTOCOL_TCP,
                                  NULL,
@@ -775,55 +879,30 @@ main (int argc, char *argv[])
       exit (1);
     }
 
-  g_signal_connect (service,
+  g_signal_connect (socket_service,
                     "incoming", G_CALLBACK (incoming_callback),
                     NULL);
 
-  clients = g_hash_table_new_full (g_int64_hash, g_int64_equal,
-                                   NULL, (GDestroyNotify) client_free);
-
-  loop = g_main_loop_new (NULL, TRUE);
-#ifdef G_OS_UNIX
-  open_mux_path ("/dev/virtio-ports/org.spice-space.webdav.0");
-#else
-  open_mux_path ("\\\\.\\Global\\org.spice-space.webdav.0");
-#endif
-
-
-  /* listen on port for incoming clients, multiplex there input into
-     virtio path, demultiplex input from there to the respective
-     clients */
-
-  g_socket_service_start (service);
-
-#ifdef WITH_AVAHI
-  mdns_client = ga_client_new (GA_CLIENT_FLAG_NO_FLAGS);
-  g_signal_connect (mdns_client, "state-changed", G_CALLBACK (mdns_state_changed), NULL);
-  if (!ga_client_start (mdns_client, &error))
-    {
-      g_printerr ("%s\n", error->message);
-      exit (1);
-    }
-#endif
-
 #ifdef G_OS_WIN32
   SERVICE_TABLE_ENTRY service_table[] =
     {
       { (char *)"spice-webdavd", service_main }, { NULL, NULL }
     };
-  if (!StartServiceCtrlDispatcher (service_table))
+  if (!getenv("DEBUG"))
     {
-      g_error ("%s", g_win32_error_message(GetLastError()));
-      exit (1);
-    }
-#else
-  g_main_loop_run (loop);
+      if (!StartServiceCtrlDispatcher (service_table))
+        {
+          g_error ("%s", g_win32_error_message (GetLastError ()));
+          exit (1);
+        }
+    } else
 #endif
+  while (!quit_service) {
+    run_service ();
+    g_usleep (G_USEC_PER_SEC);
+  }
 
-  g_main_loop_unref (loop);
-
-  output_queue_unref (mux_queue);
-  g_hash_table_unref (clients);
+  g_clear_object (&socket_service);
 
   return 0;
 }
